#include <ntddk.h>        /* required for WDM driver development */
#include <stdarg.h>     /* standard io include files */
#include <stdio.h>

#pragma optimize("", off)
#include "filter.h"    

NTSTATUS                                                            
DriverEntry                                                         
	(                                                               
	IN PDRIVER_OBJECT DriverObject,             /* pointer to the device object */
	IN PUNICODE_STRING RegistryPath             /* pointer to a unicode string representing the path */
												/*    to the drivers specific key in the registry    */
	)// ̹ ޸𸮿 óϴ 쿡 ѹ ȣ˴ϴ
{
    NTSTATUS            status = STATUS_SUCCESS;
    ULONG               ulIndex;
    PDRIVER_DISPATCH  * dispatch;

    for (ulIndex = 0, dispatch = DriverObject->MajorFunction;
         ulIndex <= IRP_MJ_MAXIMUM_FUNCTION;
         ulIndex++, dispatch++) {
        *dispatch = FilterPass;
    }

    DriverObject->MajorFunction[IRP_MJ_PNP]            = FilterDispatchPnp;
    DriverObject->MajorFunction[IRP_MJ_POWER]          = FilterDispatchPower;
    DriverObject->DriverExtension->AddDevice           = FilterAddDevice;
    DriverObject->DriverUnload                         = FilterUnload;

    return status;
}

NTSTATUS
FilterAddDevice(
    IN PDRIVER_OBJECT DriverObject,
    IN PDEVICE_OBJECT PhysicalDeviceObject
    )
{
    NTSTATUS                status = STATUS_SUCCESS;
    PDEVICE_OBJECT          deviceObject = NULL;
    PDEVICE_EXTENSION       deviceExtension;
    ULONG                   deviceType = FILE_DEVICE_UNKNOWN;

    status = IoCreateDevice (DriverObject,
                             sizeof (DEVICE_EXTENSION),
                             NULL,  // No Name
                             deviceType,
                             FILE_DEVICE_SECURE_OPEN,
                             FALSE,
                             &deviceObject);

    if (!NT_SUCCESS (status)) {
        return status;
    }

    deviceExtension = (PDEVICE_EXTENSION) deviceObject->DeviceExtension;

    deviceExtension->NextLowerDriver = IoAttachDeviceToDeviceStack (
                                       deviceObject,
                                       PhysicalDeviceObject);
    if(NULL == deviceExtension->NextLowerDriver) {

        IoDeleteDevice(deviceObject);
        return STATUS_UNSUCCESSFUL;
    }

    deviceObject->Flags |= deviceExtension->NextLowerDriver->Flags &
                            (DO_BUFFERED_IO | DO_DIRECT_IO |
                            DO_POWER_PAGABLE );

    deviceObject->DeviceType = deviceExtension->NextLowerDriver->DeviceType;

    deviceObject->Characteristics =
                          deviceExtension->NextLowerDriver->Characteristics;

    deviceExtension->Self = deviceObject;

	deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

    return STATUS_SUCCESS;

}

NTSTATUS
FilterPass (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    PDEVICE_EXTENSION           deviceExtension;

    deviceExtension = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;

    IoSkipCurrentIrpStackLocation (Irp);
    return IoCallDriver (deviceExtension->NextLowerDriver, Irp);
}

VOID                                                                
FilterUnload                                                       
	(                                                               
	IN PDRIVER_OBJECT DriverObject  /* pointer to device object */
	)
{
}

NTSTATUS
FilterDispatchPnp
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to device object */
	IN PIRP Irp                         /* pointer to an I/O request packet */
	)
{// PNPManager  IRP óϴ ó
	PIO_STACK_LOCATION pStack;                /* pointer to IRP stack */
	PDEVICE_EXTENSION deviceExtension;          /* pointer to device extention */
	NTSTATUS returnStatus = STATUS_SUCCESS;     /* stores status to be returned */
	PDEVICE_OBJECT NextLayerDeviceObject;           /* pointer to the stack device object */
	
	pStack = IoGetCurrentIrpStackLocation ( Irp );
	deviceExtension = DeviceObject->DeviceExtension;
	NextLayerDeviceObject = deviceExtension->NextLowerDriver;
			
	switch ( pStack->MinorFunction )
	{
		case IRP_MN_REMOVE_DEVICE :
		{ // IRP DeviceStack ü϶ û ǹ̷ ޵˴ϴ
			IoDetachDevice ( NextLayerDeviceObject );// Next Layer Driver  켱 ϴ
													 
			IoDeleteDevice ( DeviceObject );		 // 츮 DeviceObject մϴ
													 
			Irp->IoStatus.Status = STATUS_SUCCESS;
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;

		default:
		{
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;
	} 
	return returnStatus;
}

NTSTATUS                                                            
FilterDispatchPower
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// POWERManager  IRP óϴ ó
	PDEVICE_EXTENSION		deviceExtension;
	NTSTATUS returnStatus;
	PDEVICE_OBJECT NextLayerDeviceObject;           /* pointer to the stack device object */
	
	deviceExtension = DeviceObject->DeviceExtension;
	NextLayerDeviceObject = deviceExtension->NextLowerDriver;

	PoStartNextPowerIrp( Irp ); //  ٸ Power IRP   ִٰ PowerManager ˸ϴ
	IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
	returnStatus = PoCallDriver(NextLayerDeviceObject, Irp); // Power IRP ٸ IRP ޸, ݵ PoCallDriver()Լ ؾ մϴ

	return returnStatus;
} 

