#pragma optimize("", off)

#include <wdm.h>        
#include <stdarg.h>     
#include <stdio.h>
#include "ntddser.h"
#include "devioctl.h"
#include "VSerial.h"     
#include "Queue.h"     
#include "RS232.h"     
#include "util.h"     

NTSTATUS                                                            
DriverEntry                                                         
(                                                               
	IN PDRIVER_OBJECT DriverObject,             
	IN PUNICODE_STRING RegistryPath            
)
{
	NTSTATUS returnStatus = STATUS_SUCCESS; 

	DriverObject->MajorFunction[IRP_MJ_CREATE] = VSERIAL_Create;
	DriverObject->MajorFunction[IRP_MJ_CLOSE] = VSERIAL_Close;
	DriverObject->MajorFunction[IRP_MJ_CLEANUP] = VSERIAL_Cleanup;
	DriverObject->MajorFunction[IRP_MJ_WRITE] = VSERIAL_Write;
	DriverObject->MajorFunction[IRP_MJ_READ] = VSERIAL_Read;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = VSERIAL_DeviceIoControl;
	DriverObject->MajorFunction[IRP_MJ_PNP] = VSERIAL_Dispatch;
	DriverObject->MajorFunction[IRP_MJ_POWER] = VSERIAL_Power;
	
	DriverObject->DriverUnload = VSERIAL_Unload;
	DriverObject->DriverExtension->AddDevice = VSERIAL_AddDevice;
	return returnStatus;
}

NTSTATUS
VSERIAL_AddDevice
	(
	IN PDRIVER_OBJECT DriverObject,         
	IN PDEVICE_OBJECT PhysicalDeviceObject  
	)
{
	PDEVICE_EXTENSION deviceExtension;              
	NTSTATUS returnStatus = STATUS_SUCCESS;         
	PDEVICE_OBJECT DeviceObject = NULL;             

	returnStatus = IoCreateDevice                                       
					(                                               
						DriverObject,                               
						sizeof ( DEVICE_EXTENSION ),                
						NULL,                   
						FILE_DEVICE_SERIAL_PORT,
						0,                                          
						FALSE,                                      
						&DeviceObject       
					);

	DeviceObject->Flags &= ~DO_DEVICE_INITIALIZING;
	DeviceObject->Flags |= DO_BUFFERED_IO;

	deviceExtension = DeviceObject->DeviceExtension;

	deviceExtension->StackDeviceObject =
			IoAttachDeviceToDeviceStack
				(
					DeviceObject,
					PhysicalDeviceObject
				);
	if( !deviceExtension->StackDeviceObject )
	{
		IoDeleteDevice( DeviceObject );
		return STATUS_UNSUCCESSFUL;
	}

	deviceExtension->WaitQTail = deviceExtension->WaitQHead = 0;
	deviceExtension->pHajeQueue = AllocateQueue();
	deviceExtension->dwBaudRate = 9600;
	deviceExtension->slc.StopBits = 0;
	deviceExtension->slc.Parity = 0;
	deviceExtension->slc.WordLength = 8;
	deviceExtension->PhysicalDeviceObject = PhysicalDeviceObject;
	deviceExtension->DeviceObject = DeviceObject;
	deviceExtension->MODEMSTATUSREGISTER = 0;
	deviceExtension->MODEMCONTROLREGISTER = 0;
	deviceExtension->RemoveNotificaionIrp = 0;

	KeInitializeEvent( &deviceExtension->EmulationPollingExitEvent, NotificationEvent, FALSE );
	KeInitializeEvent( &deviceExtension->EmulationPollingAnswerEvent, NotificationEvent, FALSE );
	KeInitializeEvent( &deviceExtension->ReadPollingExitEvent, NotificationEvent, FALSE );
	KeInitializeEvent( &deviceExtension->ReadPollingAnswerEvent, NotificationEvent, FALSE );

       KeInitializeSpinLock(&deviceExtension->SpinLock);
	IoRegisterDeviceInterface( PhysicalDeviceObject, &GUID_DEVINTERFACE_COMPORT, NULL, &deviceExtension->UnicodeString );
    
	return returnStatus;
}

VOID                                                                
VSERIAL_Unload                                                       
	(                                                               
	IN PDRIVER_OBJECT DriverObject  /* pointer to device object */
	)
{
	// None
}

NTSTATUS
VSERIAL_DeferIrpCompletion
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to a device object */
	IN PIRP Irp,                        /* pointer to a I/O request packet */
	IN PVOID Context                    /* driver defined context */
	)
{
	PKEVENT event = Context;    /* kernel event passed in Context */
	KeSetEvent(event,
			   1,
			   FALSE);

	return STATUS_MORE_PROCESSING_REQUIRED;
}

void WaitIrpCancelRoutine( PDEVICE_OBJECT DeviceObject, PIRP Irp )
{
	PDEVICE_EXTENSION deviceExtension;
	PIRP PendingWaitIrp;
	int i;
	int count;

	deviceExtension = DeviceObject->DeviceExtension;
	IoSetCancelRoutine( Irp, 0 );

	count = SizeofPendingWaitIrp( deviceExtension );
	for( i = 0 ; i < count ; i ++ )
	{
		PendingWaitIrp = OutputPendingWaitIrp( deviceExtension );
		if( PendingWaitIrp == Irp )
			break;
		InputPendingWaitIrp( deviceExtension, PendingWaitIrp );
	}

	IoReleaseCancelSpinLock( Irp->CancelIrql );

	Irp->IoStatus.Status = STATUS_CANCELLED;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
}

NTSTATUS
VSERIAL_Power
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to device object */
	IN PIRP Irp                         /* pointer to an I/O request packet */
	)
{
	PDEVICE_EXTENSION pDE;

	pDE = DeviceObject->DeviceExtension;
	
	PoStartNextPowerIrp( Irp );
	IoSkipCurrentIrpStackLocation( Irp );
	return PoCallDriver( pDE->StackDeviceObject, Irp );
}

NTSTATUS
VSERIAL_Dispatch
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to device object */
	IN PIRP Irp                         /* pointer to an I/O request packet */
	)
{
	PIO_STACK_LOCATION irpStack;                /* pointer to IRP stack */
	PDEVICE_EXTENSION deviceExtension;          /* pointer to device extention */
	NTSTATUS returnStatus = STATUS_SUCCESS;     /* stores status to be returned */
	PDEVICE_OBJECT stackDeviceObject;           /* pointer to the stack device object */
	PIRP PendingWaitIrp;

	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;

	irpStack = IoGetCurrentIrpStackLocation ( Irp );
	deviceExtension = DeviceObject->DeviceExtension;
	stackDeviceObject = deviceExtension->StackDeviceObject;
			
	switch ( irpStack->MinorFunction )
	{
		case IRP_MN_START_DEVICE :
		{
			KEVENT event;   /* event used to wait for an operation to complete */
			KeInitializeEvent(&event, NotificationEvent, FALSE);
			IoCopyCurrentIrpStackLocationToNext(Irp);  
			IoSetCompletionRoutine(Irp,
								   VSERIAL_DeferIrpCompletion,
								   &event,
								   TRUE,
								   TRUE,
								   TRUE);
			returnStatus = IoCallDriver(stackDeviceObject,
											Irp);
			if (returnStatus == STATUS_PENDING) 
			{
				KeWaitForSingleObject(
					&event,
					Executive,
					KernelMode,
					FALSE,
					NULL);
			}				
			returnStatus = Irp->IoStatus.Status;
			IoCompleteRequest (Irp, IO_NO_INCREMENT);
			if(NT_SUCCESS(returnStatus))
			{
				returnStatus = RegisterSerialClassDriver( DeviceObject );
			}
		}
		break;

		case IRP_MN_SURPRISE_REMOVAL :
		{
			PIRP PendingWaitIrp;
			BOOLEAN fChanged;

			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, FALSE );
			if( deviceExtension->RemoveNotificaionIrp )
			{
				deviceExtension->RemoveNotificaionIrp->IoStatus.Status = STATUS_SUCCESS;
				IoCompleteRequest( deviceExtension->RemoveNotificaionIrp, IO_NO_INCREMENT );
				deviceExtension->RemoveNotificaionIrp = 0;
			}

			RS232_Emulation( deviceExtension, ~RTS );
			RS232_Emulation( deviceExtension, ~DTR );

			KeSetEvent( &deviceExtension->ReadPollingAnswerEvent, 0, FALSE );
			KeSetEvent( &deviceExtension->EmulationPollingAnswerEvent, 0, FALSE );

			IoCopyCurrentIrpStackLocationToNext(Irp);
			returnStatus = IoCallDriver(stackDeviceObject, Irp);                   
		}
		break;

		case IRP_MN_REMOVE_DEVICE :
		{
			UNICODE_STRING linkName;
			ULONG dwRet;
			WCHAR PropertyBuffer[255];

			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, FALSE );
			RtlFreeUnicodeString( &deviceExtension->UnicodeString );

			RtlInitUnicodeString( &linkName, deviceExtension->DOS_NAME );
			IoDeleteSymbolicLink( &linkName );

			RtlInitUnicodeString( &linkName, deviceExtension->DOS_NAME2 );
			IoDeleteSymbolicLink( &linkName );

			IoGetDeviceProperty( deviceExtension->PhysicalDeviceObject, DevicePropertyPhysicalDeviceObjectName 
				, 255, PropertyBuffer, &dwRet );

			RtlDeleteRegistryValue( RTL_REGISTRY_DEVICEMAP, L"\\SerialComm", PropertyBuffer );
			
			if(deviceExtension->pHajeQueue)
				FreeQueue(deviceExtension->pHajeQueue);
			
			IoDetachDevice ( deviceExtension->StackDeviceObject );
			IoDeleteDevice ( DeviceObject );
			IoCopyCurrentIrpStackLocationToNext(Irp);
			returnStatus = IoCallDriver(stackDeviceObject, Irp);                   
		}
		break;

		default:
		{
			IoCopyCurrentIrpStackLocationToNext(Irp);
			returnStatus = IoCallDriver(stackDeviceObject,
											Irp);
		}
		break;
	} 

	return returnStatus;
}

NTSTATUS                                                            
VSERIAL_Create
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{
	USHORT					usFeature;
	PDEVICE_EXTENSION		deviceExtension;
	HANDLE					ThreadHandle;
	NTSTATUS returnStatus = STATUS_SUCCESS;

	deviceExtension = DeviceObject->DeviceExtension;

	RS232_Emulation( deviceExtension, RTS );

	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;

	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	KeResetEvent( &deviceExtension->EmulationPollingExitEvent );
	KeResetEvent( &deviceExtension->EmulationPollingAnswerEvent );
	KeResetEvent( &deviceExtension->ReadPollingExitEvent );
	KeResetEvent( &deviceExtension->ReadPollingAnswerEvent );

	PsCreateSystemThread( &ThreadHandle, 0L, 0, 0, 0, pEmulationPollingFunction, deviceExtension ); // Emuló
	PsCreateSystemThread( &ThreadHandle, 0L, 0, 0, 0, pReadPollingFunction, deviceExtension ); // ReadIRPó
	
	return returnStatus;
}

void pReadPollingFunction(
		IN PVOID context
	)
{
	PIRP PendingWaitIrp, PendingReadIrp;
	PDEVICE_EXTENSION deviceExtension;
	PHYSICAL_ADDRESS dwInterval;
	NTSTATUS returnStatus;
	PIO_STACK_LOCATION IrpStack;
	ULONG dwRet;
	PIRP TempIrp;
	KIRQL Irql;

	KeSetPriorityThread( KeGetCurrentThread(), LOW_REALTIME_PRIORITY );
	deviceExtension = (PDEVICE_EXTENSION)context;
	dwInterval = RtlConvertLongToLargeInteger( -10000 );

	while( 1 )
	{
		returnStatus = KeWaitForSingleObject( &deviceExtension->ReadPollingExitEvent, Executive, KernelMode, FALSE, &dwInterval );
		if( returnStatus == STATUS_SUCCESS )
		{
			KeSetEvent( &deviceExtension->ReadPollingAnswerEvent, 0, FALSE );
		       KeAcquireSpinLock(&deviceExtension->SpinLock, &Irql );

			PendingWaitIrp = OutputPendingWaitIrp( deviceExtension );
			if( PendingWaitIrp )
			{
				PendingWaitIrp->IoStatus.Status = STATUS_SUCCESS;
				PendingWaitIrp->IoStatus.Information = sizeof(ULONG);
				IoSetCancelRoutine( PendingWaitIrp, 0 );
				IoCompleteRequest ( PendingWaitIrp, IO_NO_INCREMENT );
			}

			if( deviceExtension->PendingRead )
			{
				PendingReadIrp = deviceExtension->PendingRead;
				deviceExtension->PendingRead = 0;
				PendingReadIrp->IoStatus.Status = STATUS_SUCCESS;
				PendingReadIrp->IoStatus.Information = 0;
				IoSetCancelRoutine( PendingReadIrp, 0 );
				IoCompleteRequest ( PendingReadIrp, IO_NO_INCREMENT );
			}
			
			KeReleaseSpinLock(&deviceExtension->SpinLock, Irql );
			
			return;
		}

	        KeAcquireSpinLock(&deviceExtension->SpinLock, &Irql );
		{
			PendingWaitIrp = OutputPendingWaitIrp( deviceExtension );
			if( PendingWaitIrp )
			{
				if( SizeofQueue(deviceExtension->pHajeQueue) != 0 )
				{
					*(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer) = 0;
					if( deviceExtension->Wm & SERIAL_EV_RXCHAR)
					{
						*(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer) |= SERIAL_EV_RXCHAR;
					}
					if( *(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer) )
					{
						PendingWaitIrp->IoStatus.Status = STATUS_SUCCESS;
						PendingWaitIrp->IoStatus.Information = sizeof(ULONG);
						IoSetCancelRoutine( PendingWaitIrp, 0 );
						IoCompleteRequest ( PendingWaitIrp, IO_NO_INCREMENT );
					}
					else
						InputPendingWaitIrp(deviceExtension, PendingWaitIrp );
				}
				else
					InputPendingWaitIrp(deviceExtension, PendingWaitIrp );
			}
		}
		
		KeReleaseSpinLock(&deviceExtension->SpinLock, Irql );

		if( deviceExtension->PendingRead ) 
		{
		       KeAcquireSpinLock(&deviceExtension->SpinLock, &Irql );
			PendingReadIrp = deviceExtension->PendingRead;
			IrpStack = IoGetCurrentIrpStackLocation(PendingReadIrp);
			PendingReadIrp->IoStatus.Information = OutputQueue(deviceExtension->pHajeQueue, 
														(unsigned char *)(PendingReadIrp->AssociatedIrp.SystemBuffer),
														IrpStack->Parameters.Read.Length);
			
			if( PendingReadIrp->IoStatus.Information )
			{
				PendingReadIrp->IoStatus.Status = STATUS_SUCCESS;
				deviceExtension->PendingRead = 0;
				IoCompleteRequest(PendingReadIrp, IO_NO_INCREMENT);
			}
			KeReleaseSpinLock(&deviceExtension->SpinLock, Irql );
		}
	}
}

NTSTATUS                                                            
VSERIAL_Cleanup                                                        
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{
	PDEVICE_EXTENSION deviceExtension;
	ULONG ThreadHandle;
	PHYSICAL_ADDRESS dwInterval;
	int i, count;
	PIRP PendingWaitIrp;
	NTSTATUS returnStatus=STATUS_SUCCESS;

	deviceExtension = DeviceObject->DeviceExtension;

	count = SizeofPendingWaitIrp( deviceExtension );
	for( i = 0 ; i < count ; i ++ )
	{
		PendingWaitIrp = OutputPendingWaitIrp( deviceExtension );
		IoSetCancelRoutine( PendingWaitIrp, 0 );
		PendingWaitIrp->IoStatus.Status = STATUS_CANCELLED;
		PendingWaitIrp->IoStatus.Information = 0;
		IoCompleteRequest( PendingWaitIrp, IO_NO_INCREMENT );
	}

	KeSetEvent( &deviceExtension->ReadPollingExitEvent, 0, FALSE );
	KeSetEvent( &deviceExtension->EmulationPollingExitEvent, 0, FALSE );

	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return returnStatus;
}

NTSTATUS                                                            
VSERIAL_Close                                                        
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{
	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest (Irp, IO_NO_INCREMENT);
	return STATUS_SUCCESS;
}

NTSTATUS                                                            
VSERIAL_Read
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{
	PDEVICE_EXTENSION deviceExtension;
	PIO_STACK_LOCATION IrpStack;
	unsigned char *pBuffer;
	ULONG dwSize;
	NTSTATUS returnStatus=STATUS_SUCCESS;
	KIRQL OldIrql;

	deviceExtension = DeviceObject->DeviceExtension;
       KeAcquireSpinLock(&deviceExtension->SpinLock, &OldIrql );

	IrpStack = IoGetCurrentIrpStackLocation ( Irp );
	Irp->IoStatus.Status = returnStatus;

	Irp->IoStatus.Information = OutputQueue(deviceExtension->pHajeQueue,
												Irp->AssociatedIrp.SystemBuffer,
												IrpStack->Parameters.Read.Length);
	
	
	if( Irp->IoStatus.Information ) 
	{
			KeReleaseSpinLock(&deviceExtension->SpinLock, OldIrql );
			IoCompleteRequest( Irp, IO_NO_INCREMENT );
			return STATUS_SUCCESS;
	}

	IoMarkIrpPending( Irp );
	deviceExtension->PendingRead = Irp;
	KeReleaseSpinLock(&deviceExtension->SpinLock, OldIrql );
	return STATUS_PENDING;
}

NTSTATUS                                                            
VSERIAL_Write
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{
	PDEVICE_EXTENSION deviceExtension;
	PIO_STACK_LOCATION pStack;
	KIRQL Irql;

	deviceExtension = DeviceObject->DeviceExtension;

	pStack = IoGetCurrentIrpStackLocation( Irp );
	
	// Emulation... TX ---> RX
	Irp->IoStatus.Information = InputQueue(deviceExtension->pHajeQueue,
										Irp->AssociatedIrp.SystemBuffer,
										pStack->Parameters.Write.Length);
	
	Irp->IoStatus.Status = STATUS_SUCCESS;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return STATUS_SUCCESS;
}

NTSTATUS                                                            
VSERIAL_DeviceIoControl
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{
	PIO_STACK_LOCATION irpStack, nextStack;
	ULONG ioControlCode;
	NTSTATUS returnStatus;
	PDEVICE_EXTENSION deviceExtension;
	ULONG dwI;
	PIRP PendingIrp, PendingWaitIrp;
	
	deviceExtension = DeviceObject->DeviceExtension;
	irpStack = IoGetCurrentIrpStackLocation ( Irp );
	ioControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;

	Irp->IoStatus.Status = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
	returnStatus = STATUS_SUCCESS;

	switch ( ioControlCode )
	{
		case IOCTL_SERIAL_SET_BAUD_RATE :
		{
			PSERIAL_BAUD_RATE pBuffer;
			pBuffer = (PSERIAL_BAUD_RATE) Irp->AssociatedIrp.SystemBuffer;
			deviceExtension->dwBaudRate = pBuffer->BaudRate;
			Irp->IoStatus.Information = 0;
			break;
		}
		case IOCTL_SERIAL_GET_BAUD_RATE:
		{
			PSERIAL_BAUD_RATE pBuffer;
			pBuffer = (PSERIAL_BAUD_RATE) Irp->AssociatedIrp.SystemBuffer;
			pBuffer->BaudRate = deviceExtension->dwBaudRate;
			Irp->IoStatus.Information = sizeof(SERIAL_BAUD_RATE);
			break;
		}
		case IOCTL_SERIAL_GET_MODEM_CONTROL:
		{
			ULONG * pBuffer;
			pBuffer = (ULONG *)Irp->AssociatedIrp.SystemBuffer;
			*pBuffer = deviceExtension->MODEMCONTROLREGISTER;
			Irp->IoStatus.Information = sizeof(ULONG);
			break;
		}
		case IOCTL_SERIAL_SET_MODEM_CONTROL:
		{
			ULONG * pBuffer;
			pBuffer = (ULONG *)Irp->AssociatedIrp.SystemBuffer;
			deviceExtension->MODEMCONTROLREGISTER = *pBuffer;
			if( *pBuffer & DTR )
				RS232_Emulation( deviceExtension, DTR );
			if( *pBuffer & CTS )
				RS232_Emulation(  deviceExtension,CTS );
			if( !(*pBuffer & DTR) )
				RS232_Emulation(  deviceExtension,~DTR );
			if( !(*pBuffer & DTR) )
				RS232_Emulation( deviceExtension, ~CTS );

			Irp->IoStatus.Information = sizeof(ULONG);
			break;
		}
		case IOCTL_SERIAL_SET_LINE_CONTROL:
		{
			PSERIAL_LINE_CONTROL pBuffer;
			
			pBuffer = (PSERIAL_LINE_CONTROL)Irp->AssociatedIrp.SystemBuffer;

			if( (deviceExtension->slc.StopBits != pBuffer->StopBits) ||
				(deviceExtension->slc.Parity != pBuffer->Parity) ||
				(deviceExtension->slc.WordLength != pBuffer->WordLength)
				)
			{
				deviceExtension->slc.StopBits = pBuffer->StopBits; 
				deviceExtension->slc.Parity = pBuffer->Parity;   
				deviceExtension->slc.WordLength = pBuffer->WordLength;
			}
			break;
		}
		case IOCTL_SERIAL_GET_LINE_CONTROL:
		{
			PSERIAL_LINE_CONTROL pBuffer;
			pBuffer = (PSERIAL_LINE_CONTROL)Irp->AssociatedIrp.SystemBuffer;
			pBuffer->StopBits = 1;  
			pBuffer->Parity = 0;    
			pBuffer->WordLength = 8;    
			Irp->IoStatus.Information = sizeof(SERIAL_LINE_CONTROL);
			break;
		}
		case IOCTL_SERIAL_SET_TIMEOUTS:
		{
			PSERIAL_TIMEOUTS pBuffer;
			pBuffer = (PSERIAL_TIMEOUTS)Irp->AssociatedIrp.SystemBuffer;
			deviceExtension->stimeout.ReadIntervalTimeout = pBuffer->ReadIntervalTimeout;
			deviceExtension->stimeout.ReadTotalTimeoutMultiplier = pBuffer->ReadTotalTimeoutMultiplier;
			deviceExtension->stimeout.ReadTotalTimeoutConstant = pBuffer->ReadTotalTimeoutConstant;
			deviceExtension->stimeout.WriteTotalTimeoutMultiplier = pBuffer->WriteTotalTimeoutMultiplier;
			deviceExtension->stimeout.WriteTotalTimeoutConstant = pBuffer->WriteTotalTimeoutConstant;
			break;
		}
		case IOCTL_SERIAL_GET_TIMEOUTS:
		{
			PSERIAL_TIMEOUTS pBuffer;
			pBuffer = (PSERIAL_TIMEOUTS)Irp->AssociatedIrp.SystemBuffer;
			pBuffer->ReadIntervalTimeout = deviceExtension->stimeout.ReadIntervalTimeout;
			pBuffer->ReadTotalTimeoutMultiplier = deviceExtension->stimeout.ReadTotalTimeoutMultiplier;
			pBuffer->ReadTotalTimeoutConstant = deviceExtension->stimeout.ReadTotalTimeoutConstant;
			pBuffer->WriteTotalTimeoutMultiplier = deviceExtension->stimeout.WriteTotalTimeoutMultiplier;
			pBuffer->WriteTotalTimeoutConstant = deviceExtension->stimeout.WriteTotalTimeoutConstant;
			Irp->IoStatus.Information = sizeof(SERIAL_TIMEOUTS);
			break;
		}
		case IOCTL_SERIAL_SET_DTR:
			RS232_Emulation(  deviceExtension,DTR );
			break;
		case IOCTL_SERIAL_CLR_DTR:
			RS232_Emulation(  deviceExtension,~DTR );
			break;
		case IOCTL_SERIAL_SET_RTS:
			RS232_Emulation(  deviceExtension,RTS );
			break;
		case IOCTL_SERIAL_CLR_RTS:
			RS232_Emulation(  deviceExtension,~RTS );
			break;
		case IOCTL_SERIAL_GET_WAIT_MASK:
		{
			ULONG *pBuffer;
			pBuffer = (ULONG *)Irp->AssociatedIrp.SystemBuffer;
			*pBuffer = deviceExtension->Wm;
			Irp->IoStatus.Information = sizeof(ULONG);
			break;
		}
		case IOCTL_SERIAL_SET_WAIT_MASK:
		{
			ULONG *pBuffer;
			ULONG WaitIrps;
		       KIRQL OldIrql;

			pBuffer = (ULONG *)Irp->AssociatedIrp.SystemBuffer;

		       KeAcquireSpinLock(&deviceExtension->SpinLock, &OldIrql );

			PendingWaitIrp = OutputPendingWaitIrp(deviceExtension );
			if( PendingWaitIrp )
			{
				*(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer) = 0;
				PendingWaitIrp->IoStatus.Status = STATUS_SUCCESS;
				PendingWaitIrp->IoStatus.Information = sizeof(ULONG);
				IoSetCancelRoutine( PendingWaitIrp, 0 );
				IoCompleteRequest ( PendingWaitIrp, IO_NO_INCREMENT );
			}

		       KeReleaseSpinLock(&deviceExtension->SpinLock, OldIrql );

			deviceExtension->Wm |= (ULONG)(*pBuffer);
			if( (ULONG)(*pBuffer) == 0x0 )
				deviceExtension->Wm = 0;
			break;
		}
		case IOCTL_SERIAL_WAIT_ON_MASK:
		{
		        KIRQL OldIrql;
			*(ULONG *)(Irp->AssociatedIrp.SystemBuffer) = 0;

		       KeAcquireSpinLock(&deviceExtension->SpinLock, &OldIrql );

			returnStatus = STATUS_PENDING;
			IoMarkIrpPending(Irp);
			InputPendingWaitIrp(deviceExtension, Irp );
			IoSetCancelRoutine( Irp, WaitIrpCancelRoutine );

		       KeReleaseSpinLock(&deviceExtension->SpinLock, OldIrql );
			return STATUS_PENDING;
		}
		case IOCTL_SERIAL_PURGE:
		{
		    KIRQL OldIrql;

		    ULONG *pBuffer;
			pBuffer = (ULONG *)Irp->AssociatedIrp.SystemBuffer;
			
		    KeAcquireSpinLock(&deviceExtension->SpinLock, &OldIrql );

			if( *(pBuffer) & SERIAL_PURGE_RXABORT)
			{
				PIRP PendingIrp;
				PendingIrp = deviceExtension->PendingRead;

				if( PendingIrp == 0 )
				{
				    KeReleaseSpinLock(&deviceExtension->SpinLock, OldIrql );
				}
				else
				{
					PendingIrp->IoStatus.Status = STATUS_CANCELLED;	
					PendingIrp->IoStatus.Information = 0;
					deviceExtension->PendingRead = 0;
				    KeReleaseSpinLock(&deviceExtension->SpinLock, OldIrql );
					IoCompleteRequest( PendingIrp, IO_NO_INCREMENT );
				}
			}
			else
			    KeReleaseSpinLock(&deviceExtension->SpinLock, OldIrql );

			if( *(pBuffer) & SERIAL_PURGE_RXCLEAR)
			{
				ResetQueue(deviceExtension->pHajeQueue);
			}

			if( *(pBuffer) & SERIAL_PURGE_TXCLEAR)
			{
			}

			Irp->IoStatus.Information = sizeof(ULONG);
			break;
		}
		case IOCTL_SERIAL_GET_MODEMSTATUS:
		{
			unsigned char *pBuffer;
			pBuffer = (unsigned char *)Irp->AssociatedIrp.SystemBuffer;
			*pBuffer = (unsigned char)deviceExtension->MODEMSTATUSREGISTER;
			*pBuffer |= CTS;
						
			deviceExtension->MODEMSTATUSREGISTER &= ~TDCD;
			deviceExtension->MODEMSTATUSREGISTER &= ~TDSR;
			deviceExtension->MODEMSTATUSREGISTER &= ~TCTS;
			Irp->IoStatus.Information = sizeof(ULONG);
			break;
		}
		case IOCTL_SERIAL_GET_COMMSTATUS:
		{
			PSERIAL_STATUS pBuffer;
			pBuffer = (PSERIAL_STATUS)Irp->AssociatedIrp.SystemBuffer;
			pBuffer->Errors = 0;
			pBuffer->HoldReasons = 0;
			pBuffer->AmountInInQueue = SizeofQueue( deviceExtension->pHajeQueue );
			pBuffer->AmountInOutQueue = 0;
			pBuffer->EofReceived = 0;
			pBuffer->WaitForImmediate = 0;
			Irp->IoStatus.Information = sizeof(SERIAL_STATUS);
			break;
		}
		case IOCTL_REMOVE_NOTIFICATION: 
			deviceExtension->RemoveNotificaionIrp = Irp;
			IoMarkIrpPending( Irp );
			return STATUS_PENDING;
	}  
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return returnStatus;                       
}

