#include <ntddk.h>        /* required for WDM driver development */
#include <stdarg.h>     /* standard io include files */
#include <stdio.h>

#include "devioctl.h"
#include <initguid.h>

DEFINE_GUID(GUID_SAMPLE, 
0x5665dec0, 0xa40a, 0x11d1, 0xb9, 0x84, 0x0, 0x20, 0xaf, 0xd7, 0x97, 0x78);
//  츮  DeviceStack  Interface GUID մϴ

#pragma optimize("", off)
#include "SAMPLE4.h"    
#include "createmethod.h"

void IrpStringOut( PIRP Irp ) // IRP Ʈմϴ
{
	KIRQL Irql;
	PIO_STACK_LOCATION pStack;
	
	pStack = IoGetCurrentIrpStackLocation( Irp );

	switch( pStack->MajorFunction )
	{
	case IRP_MJ_CREATE:
		DbgPrint("[IRP_MJ_CREATE]");
		break;
	case IRP_MJ_READ:
		DbgPrint("[IRP_MJ_READ]");
		break;
	case IRP_MJ_WRITE:
		DbgPrint("[IRP_MJ_WRITE]");
		break;
	case IRP_MJ_CLEANUP:
		DbgPrint("[IRP_MJ_CLEANUP]");
		break;
	case IRP_MJ_CLOSE:
		DbgPrint("[IRP_MJ_CLOSE]");
		break;
	case IRP_MJ_DEVICE_CONTROL:
		DbgPrint("[IRP_MJ_DEVICE_CONTROL]");
		DbgPrint("ControlCode = %8X", pStack->Parameters.DeviceIoControl.IoControlCode);
		break;
	case IRP_MJ_PNP:
		DbgPrint("[IRP_MJ_PNP]");
		switch( pStack->MinorFunction )
		{
			case IRP_MN_START_DEVICE:
				DbgPrint("IRP_MN_START_DEVICE");
				break;
			case IRP_MN_REMOVE_DEVICE:
				DbgPrint("IRP_MN_REMOVE_DEVICE");
				break;
			case IRP_MN_SURPRISE_REMOVAL:
				DbgPrint("IRP_MN_SURPRISE_REMOVAL");
				break;
			default:
				DbgPrint("ANY MN FUNCTION");
				break;
		}
		break;
	case IRP_MJ_POWER:
		DbgPrint("[IRP_MJ_POWER]");
		switch( pStack->MinorFunction )
		{
			case IRP_MN_SET_POWER:
				DbgPrint("IRP_MN_SET_POWER");
				break;
			case IRP_MN_QUERY_POWER:
				DbgPrint("IRP_MN_QUERY_POWER");
				break;
			default:
				DbgPrint("ANY MN FUNCTION");
				break;
		}
		break;
	default:
		DbgPrint("[ANY IRP] = 0x%8X\n", Irp );
		return;
	}
	DbgPrint("\n");
}

NTSTATUS                                                            
DriverEntry                                                         
	(                                                               
	IN PDRIVER_OBJECT DriverObject,             /* pointer to the device object */
	IN PUNICODE_STRING RegistryPath             /* pointer to a unicode string representing the path */
												/*    to the drivers specific key in the registry    */
	)// ̹ ޸𸮿 óϴ 쿡 ѹ ȣ˴ϴ
{
	NTSTATUS returnStatus = STATUS_SUCCESS;     /* stores status to be returned */

	DbgPrint("DriverEntry ++ \n" );

	DriverObject->DriverUnload = SAMPLE4_Unload;
	DriverObject->DriverExtension->AddDevice = SAMPLE4_AddDevice;
	DriverObject->MajorFunction[IRP_MJ_PNP] = SAMPLE4_PnpDispatch;
	DriverObject->MajorFunction[IRP_MJ_POWER] = SAMPLE4_PowerDispatch;
	DriverObject->MajorFunction[IRP_MJ_CREATE] = SAMPLE4_Create;
	DriverObject->MajorFunction[IRP_MJ_CLOSE] = SAMPLE4_Close;
	DriverObject->MajorFunction[IRP_MJ_READ] = SAMPLE4_Read;
	// Win32 API ReadFile();
	DriverObject->MajorFunction[IRP_MJ_WRITE] = SAMPLE4_Write;
	// Win32 API WriteFile();
	DriverObject->MajorFunction[IRP_MJ_CLEANUP] = SAMPLE4_Cleanup;
	// Win32 API CloseHandle();

	DbgPrint("DriverEntry -- \n" );
	return returnStatus;
}

NTSTATUS
SAMPLE4_AddDevice
	(
	IN PDRIVER_OBJECT DriverObject,         
	IN PDEVICE_OBJECT PhysicalDeviceObject  
	)// PNPManager ̰ ؼ IRP DeviceStack ϶ û մϴ
{
	/* IRP DeviceStack ϱ ؼ  ܰ踦 ġ ˴ϴ 
	1. DeviceObject۾
	2. ϴ DeviceStack DeviceObject ø
	3. DeviceObject Flag ϱ
	*/
	PDEVICE_EXTENSION deviceExtension;              
	NTSTATUS returnStatus = STATUS_SUCCESS;         
	PDEVICE_OBJECT DeviceObject = NULL;             
	ULONG dwRet;

	DbgPrint("SAMPLE4_AddDevice ++ \n" );

	returnStatus = IoCreateDevice                                       
					(                                               
						DriverObject,                               
						sizeof ( DEVICE_EXTENSION ),                
						NULL,                   
						FILE_DEVICE_UNKNOWN,
						FILE_AUTOGENERATED_DEVICE_NAME,                                          
						FALSE,                                      
						&DeviceObject       
					);// DeviceObject մϴ

	if( !NT_SUCCESS( returnStatus ) )
	{
		DbgPrint("IoCreateDevice() Error, Status = %8X\n", returnStatus );
		goto SAMPLE4_AddDevice_Exit;
	}

	deviceExtension = DeviceObject->DeviceExtension;

	RtlZeroMemory( deviceExtension, sizeof(DEVICE_EXTENSION));

	deviceExtension->PhysicalDeviceObject = PhysicalDeviceObject;
	
	deviceExtension->DeviceObject = DeviceObject;

	deviceExtension->NextLayerDeviceObject =
	IoAttachDeviceToDeviceStack (
							DeviceObject,
							PhysicalDeviceObject	
	);// ϴ DeviceStack DeviceObject øϴ

	DeviceObject->Flags |= deviceExtension->NextLayerDeviceObject->Flags & 
							( DO_POWER_PAGABLE  | DO_POWER_INRUSH); // DeviceObject Flag մϴ

	DeviceObject->Flags &= ~DO_DEVICE_INITIALIZING; // DeviceObject Flag մϴ

	DeviceObject->Flags |= DO_BUFFERED_IO; // IRP SystemBuffer ؼ α׷ۿ ɰ Ͻ

	IoRegisterDeviceInterface( PhysicalDeviceObject, &GUID_SAMPLE, NULL, &deviceExtension->UnicodeString );
	//  츮  DeviceStack  InterfaceGuid մϴ
	// ϵǴ deviceExtension->UnicodeString ڰ ϴ ̸Դϴ


SAMPLE4_AddDevice_Exit:
	DbgPrint("SAMPLE4_AddDevice -- \n" );
	return returnStatus;
}

VOID                                                                
SAMPLE4_Unload                                                       
	(                                                               
	IN PDRIVER_OBJECT DriverObject  /* pointer to device object */
	)
{// ̻ ̹  DeviceObject  ̰ ȣ, ̹ ޸𸮿 ŵ˴ϴ

	DbgPrint("SAMPLE4_Unload ++ \n" );

	// Do job..

	DbgPrint("SAMPLE4_Unload -- \n" );
}

NTSTATUS
SAMPLE4_DeferIrpCompletion
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to a device object */
	IN PIRP Irp,                        /* pointer to a I/O request packet */
	IN PVOID Context                    /* driver defined context */
	)// ɾ IRP NextLayer ̹ IoCompleteRequest()Լ ؼ ûǴ 쿡 ȣ˴ϴ
{
	PKEVENT event = Context;    /* kernel event passed in Context */

	DbgPrint("SAMPLE4_DeferIrpCompletion ++ \n" );

	KeSetEvent(event,
			   1,
			   FALSE);

	
	DbgPrint("SAMPLE4_DeferIrpCompletion -- \n" );

	return STATUS_MORE_PROCESSING_REQUIRED; // IoCompleteRequest()Լ Ͽ,  IRP  ߴ϶ û մϴ
}

NTSTATUS
SAMPLE4_PnpDispatch
	(
	IN PDEVICE_OBJECT DeviceObject,     /* pointer to device object */
	IN PIRP Irp                         /* pointer to an I/O request packet */
	)
{// PNPManager  IRP óϴ ó
	PIO_STACK_LOCATION pStack;                /* pointer to IRP stack */
	PDEVICE_EXTENSION deviceExtension;          /* pointer to device extention */
	NTSTATUS returnStatus = STATUS_SUCCESS;     /* stores status to be returned */
	PDEVICE_OBJECT NextLayerDeviceObject;           /* pointer to the stack device object */
	
	DbgPrint("SAMPLE4_PnpDispatch ++ \n" );

	IrpStringOut( Irp );

	pStack = IoGetCurrentIrpStackLocation ( Irp );
	deviceExtension = DeviceObject->DeviceExtension;
	NextLayerDeviceObject = deviceExtension->NextLayerDeviceObject;
			
	switch ( pStack->MinorFunction )
	{
		case IRP_MN_START_DEVICE :
		{// DeviceStack ϶ 㰡 ǹմϴ
			KEVENT event;   /* event used to wait for an operation to complete */
			KeInitializeEvent(&event, NotificationEvent, FALSE);
			IoCopyCurrentIrpStackLocationToNext(Irp);  
			IoSetCompletionRoutine(Irp,
								   SAMPLE4_DeferIrpCompletion,
								   &event,
								   TRUE,
								   TRUE,
								   TRUE); // SAMPLE4_DeferIrpCompletion()Լ  IRP  Next Layer Driver ȣϴ
										  // IoCompleteRequest()Լ ؼ ȣǵ մϴ
			
			returnStatus = IoCallDriver(NextLayerDeviceObject,Irp); // NextLayer Driver   ѵ,  Ȯմϴ
			if (returnStatus == STATUS_PENDING) 
			{
				KeWaitForSingleObject(
					&event,
					Executive,
					KernelMode,
					FALSE,
					NULL);
				returnStatus = Irp->IoStatus.Status;
			}
			if( !NT_SUCCESS( returnStatus ) )
			{
				DbgPrint("Start Device Fail\n" );
				IoCompleteRequest( Irp, IO_NO_INCREMENT );
				goto SAMPLE4_PnpDispatch_Exit;
			}
			// NextLayer Driver    ٷ, 츮 ΰ ص ٴ ǹԴϴ

			// Do Job...
			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, TRUE ); 
			// ڰ 츮 DeviceStack ϴ  

			IoCompleteRequest( Irp, IO_NO_INCREMENT ); //  IRP ûմϴ
			// SAMPLE4_DeferIrpCompletion()Լ 츮 IRP  Next Layer Driver ȣ IoCompleteRequest() 
		}	// ϵ "STATUS_MORE_PROCESSING_REQUIRED" ױ⶧, 츮 û ؾմϴ
		break;

		case IRP_MN_SURPRISE_REMOVAL :
		{ // ڱ ġ PC ŵǾٴ ǹ̿ ޵˴ϴ

			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, FALSE );
			// ڰ 츮 DeviceStack ϴ  

			Irp->IoStatus.Status = STATUS_SUCCESS;
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;

		case IRP_MN_REMOVE_DEVICE :
		{ // IRP DeviceStack ü϶ û ǹ̷ ޵˴ϴ
			IoSetDeviceInterfaceState( &deviceExtension->UnicodeString, FALSE );
			// ڰ 츮 DeviceStack ϴ  

			RtlFreeUnicodeString( &deviceExtension->UnicodeString );
			// ڰ ϴ ̸ ޸𸮿 Ѵ

			IoDetachDevice ( NextLayerDeviceObject );// Next Layer Driver  켱 ϴ
													 // SAMPLE4_AddDevice() IoAttachDeviceToDeviceStack()Լ ݴǹ
			IoDeleteDevice ( DeviceObject );		 // 츮 DeviceObject մϴ
													 // SAMPLE4_AddDevice() IoCreateDevice()Լ ݴǹ
			Irp->IoStatus.Status = STATUS_SUCCESS;
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;

		default:
		{
			IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
			returnStatus = IoCallDriver(NextLayerDeviceObject, Irp);
		}
		break;
	} 

SAMPLE4_PnpDispatch_Exit:
	DbgPrint("SAMPLE4_PnpDispatch -- \n" );

	return returnStatus;
}

NTSTATUS                                                            
SAMPLE4_PowerDispatch
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// POWERManager  IRP óϴ ó
	PDEVICE_EXTENSION		deviceExtension;
	NTSTATUS returnStatus;
	PDEVICE_OBJECT NextLayerDeviceObject;           /* pointer to the stack device object */
	
	DbgPrint("SAMPLE4_PowerDispatch ++ \n" );

	IrpStringOut( Irp );

	deviceExtension = DeviceObject->DeviceExtension;
	NextLayerDeviceObject = deviceExtension->NextLayerDeviceObject;

	PoStartNextPowerIrp( Irp ); //  ٸ Power IRP   ִٰ PowerManager ˸ϴ
	IoSkipCurrentIrpStackLocation( Irp ); //  IRP  ɶ ̻   ʰڴٴ ǹԴϴ
	returnStatus = PoCallDriver(NextLayerDeviceObject, Irp); // Power IRP ٸ IRP ޸, ݵ PoCallDriver()Լ ؾ մϴ
	
	DbgPrint("SAMPLE4_PowerDispatch -- \n" );
	return returnStatus;
} 

NTSTATUS                                                            
SAMPLE4_Create
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  CreateFile()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION pStack;

	DbgPrint("SAMPLE4_Create ++ \n" );

	IrpStringOut( Irp );

	pStack = IoGetCurrentIrpStackLocation( Irp );

	if( !AllocateStructForFileContext( pStack->FileObject ) ) // CreateFile("̸\STRING") ȮϿ FsContext մϴ
		returnStatus = STATUS_INVALID_PARAMETER;
	
	DbgPrint("SAMPLE4_Create -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 

NTSTATUS                                                            
SAMPLE4_Close
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  CloseHandle()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION pStack;

	DbgPrint("SAMPLE4_Close ++ \n" );

	IrpStringOut( Irp );

	pStack = IoGetCurrentIrpStackLocation( Irp );

	FreeStructForFileContext( pStack->FileObject ); //  FsContext մϴ
	
	DbgPrint("SAMPLE4_Close -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 

NTSTATUS                                                            
SAMPLE4_Write
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  WriteFile()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION pStack;
	ULONG dwRequestSize;
	char * pClientBuffer;

	DbgPrint("SAMPLE4_Write ++ \n" );

	IrpStringOut( Irp );

	pStack = IoGetCurrentIrpStackLocation( Irp );

	dwRequestSize = pStack->Parameters.Write.Length;
	// 밡 ũ⸦ ȮѴ...
	pClientBuffer = Irp->AssociatedIrp.SystemBuffer; // DO_BUFFERED_IO

	switch( GetTypeCreate(pStack->FileObject) )
	{
		case STRINGTYPE:
		{
			PCONTEXT_FOR_STRING	pContextForString;
			pContextForString = (PCONTEXT_FOR_STRING)pStack->FileObject->FsContext;
			dwRequestSize = ((int)dwRequestSize > MAXSTRINGBUFFERCOUNT)?
				0:dwRequestSize;

			if( !dwRequestSize )
			{
				returnStatus = STATUS_INVALID_PARAMETER;
			}
			else
			{
				memcpy( pContextForString->StringBuffer, pClientBuffer, dwRequestSize );
				pContextForString->CurrentRemainBufferSize = dwRequestSize;
			}
		}
		break;
		default:
			break;
	}
	Irp->IoStatus.Information = dwRequestSize; //  ũ⸦ ´..
	
	DbgPrint("SAMPLE4_Write -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 

// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------
/*
	 ؼ ƾ  ҽ ִ.
	ó  .

	α׷ APP ReadFile()Լ ȣѴ. 
	-> ̹̽ IRP_MJ_READɾ ؼ  PENDING RETURNó Ѵ.
	-> ᱹ,  ٽ α׷ APP ư ȴ.

	, α׷   ó  ۾ Ѵ.
	
	ó 1) CancelIo()Լ ȣϿ,  PENDING RETURN IRP ҿûѴ. ׷, CloseHandle()Լ θ, α׷ Ѵ.
	ó 2) α׷ CloseHandle()Լ θ, α׷ Ѵ.
	ó 3) α׷ CloseHandle()Լ θ ʰ ٷ Ѵ.

	̿  ó Ѵ ϴ, ̹̽   ϱ ؼ, CancelRoutine IRP_MJ_CLEANUPóⰡ ־ Ѵ.

	ó 1 ϴ  ȣǴ   .
	α׷ -> CancelIo() -> CancelRoutine
	α׷ -> CloseHandle() -> IRP_MJ_CLEANUPó
								  -> IRP_MJ_CLOSEó (IRP_MJ_READɾ Ǿ)

	ó 2 ϴ  ȣǴ   .
	α׷ -> CloseHandle() -> IRP_MJ_CLEANUPó
								  -> IRP_MJ_CLOSEó (IRP_MJ_READɾ Ǿ)

	ó 3 ϴ  ȣǴ   .
	α׷ -> IO Manager -> CancelRoutine
						CloseHandle() -> IRP_MJ_CLEANUPó
									  -> IRP_MJ_CLOSEó (IRP_MJ_READɾ Ǿ)
*/	
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------

NTSTATUS                                                            
SAMPLE4_Read
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  ReadFile()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION pStack;
	ULONG dwRequestSize;
	char * pClientBuffer;
	PDEVICE_EXTENSION deviceExtension;

	DbgPrint("SAMPLE4_Read ++ \n" );

	IrpStringOut( Irp );

	pStack = IoGetCurrentIrpStackLocation( Irp );

	dwRequestSize = pStack->Parameters.Read.Length;
	// 밡 ũ⸦ ȮѴ... 
	pClientBuffer = Irp->AssociatedIrp.SystemBuffer; // DO_BUFFERED_IO

	switch( GetTypeCreate(pStack->FileObject) )
	{
		case STRINGTYPE: // ó , Pendingġ մϴ
			{
				PCONTEXT_FOR_STRING	pContextForString;

				pContextForString = (PCONTEXT_FOR_STRING)pStack->FileObject->FsContext;
				dwRequestSize = ((int)dwRequestSize >= pContextForString->CurrentRemainBufferSize)?
					pContextForString->CurrentRemainBufferSize:dwRequestSize;

				if( !dwRequestSize )
				{
					returnStatus = STATUS_INVALID_PARAMETER;
				}
				else
				{
					memcpy( pClientBuffer, pContextForString->StringBuffer, dwRequestSize );
				}

				InsertTailList( &pContextForString->ListHead, &(Irp->Tail.Overlay.ListEntry) );				
				//  IRP ListHead Ŵ
				// Ǵ ListHead FileObject->FsContext ִٴ   غ!
			} 

			returnStatus = STATUS_PENDING; // Pending ReturnҰ ˸
			IoMarkIrpPending( Irp );
			IoSetCancelRoutine( Irp, CancelRoutine ); // CancelRoutine մϴ
			
			DbgPrint("SAMPLE4_Read --, PENDING Return!! \n" );

			return returnStatus; // STATUS_PENDING մϴ

		default:
			break;
	}
	Irp->IoStatus.Information = dwRequestSize; //  ũ⸦ ´..
	
	DbgPrint("SAMPLE4_Read -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 

/* CancelRoutine IO Manager IoCancelIrp()Լ ؼ ȣ˴ϴ
׷, IoCancelIrp()Լ CancelRoutine() ȣϱ IoAcquireCancelSpinLock()Լ ؼ SpinLock ȹѵڿ
CancelRoutine() θ⶧, ̹ CancelRoutine()  ñ⿡ IoReleaseCancelSpinLock()Լ ҷ־ մϴ */
void CancelRoutine( PDEVICE_OBJECT DeviceObject, PIRP Irp )
{
	DbgPrint("CancelRoutine ++ \n" );

	IrpStringOut( Irp );

	RemoveEntryList( &(Irp->Tail.Overlay.ListEntry) );
	// CancelSpinLock ݳϱ , 켱 ҵ IRP ListHead ؾ Ѵ

	IoReleaseCancelSpinLock(Irp->CancelIrql); //<<<---- ʼ

	Irp->IoStatus.Status = STATUS_CANCELLED;
	Irp->IoStatus.Information = 0;
	IoCompleteRequest(Irp, IO_NO_INCREMENT); // IRP óûմϴ
	DbgPrint("IRP Cancelled!!\n" );

	DbgPrint("CancelRoutine -- \n" );
}

NTSTATUS                                                            
SAMPLE4_Cleanup
	(                                                               
		IN PDEVICE_OBJECT DeviceObject,                             
		IN PIRP Irp                                                 
	)
{// Application  CloseHandle()Լ Ǵ ó
	NTSTATUS returnStatus = STATUS_SUCCESS;
	PIO_STACK_LOCATION pStack;

	pStack = IoGetCurrentIrpStackLocation( Irp );

	DbgPrint("SAMPLE4_Cleanup ++ \n" );

	IrpStringOut( Irp );

	switch( GetTypeCreate(pStack->FileObject) )
	{
		case STRINGTYPE: 
			{
				PCONTEXT_FOR_STRING	pContextForString;
				KIRQL OldIrql;
				LIST_ENTRY RemovedIrpListHead;
				PLIST_ENTRY pTargetListEntry;
				PIRP pTargetIrp;

				pContextForString = (PCONTEXT_FOR_STRING)pStack->FileObject->FsContext;
				
/*
				IRP_MJ_CLEANUPó    Pending IRP ؾ Ѵ.

				1) CancelSpinLock 켱 ȹؾ Ѵ. ׷ , شϴ IRP õ ҷƾ ȣ  ִ
				2) ListHead(IRP_MJ_CLEANUPɾ Բ ޵ FileObject->FsContext  ListHead Ͽ,  IRP List ϳ  
				3)   IRP CancelRoutine 켱 
				4)   IRP ݵ غ ListHead(RemovedIrpListHead) Ѿ Ѵ
				5) FileObject->FsContext  ListHead   , 2)-5) ݺѴ
				6) CancelSpinLock ݳؾ Ѵ.

				7)  غ ListHead(RemovedIrpListHead)  IRP List ٽ  
				8)   IRP IoCompleteRequest()Լ Ͽ, Canceló ؾ Ѵ
				9) ListHead(RemovedIrpListHead)   , 7) 9) ݺѴ
*/
				InitializeListHead( &RemovedIrpListHead );

				IoAcquireCancelSpinLock( &OldIrql );
				while( !IsListEmpty( &pContextForString->ListHead ) )
				{
					pTargetListEntry = RemoveHeadList( &pContextForString->ListHead );
					InsertTailList( &RemovedIrpListHead, pTargetListEntry );
			        pTargetIrp = CONTAINING_RECORD (pTargetListEntry, IRP, Tail.Overlay.ListEntry);
					IoSetCancelRoutine( pTargetIrp, NULL );
				}
				IoReleaseCancelSpinLock( OldIrql );

				while( !IsListEmpty( &RemovedIrpListHead ) )
				{
					pTargetListEntry = RemoveHeadList( &RemovedIrpListHead );
			        pTargetIrp = CONTAINING_RECORD (pTargetListEntry, IRP, Tail.Overlay.ListEntry);
					pTargetIrp->IoStatus.Status = STATUS_CANCELLED;
					pTargetIrp->IoStatus.Information = 0;
					IoCompleteRequest( pTargetIrp, IO_NO_INCREMENT );
				}
			}
			break;

		default:
			break;
	}
	DbgPrint("SAMPLE4_Cleanup -- \n" );

	Irp->IoStatus.Status = returnStatus;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );

	return returnStatus;
} 
