
#include "MouseFilterDriver.h"
#pragma optimize("", off)


NTSTATUS DriverEntry( PDRIVER_OBJECT pDriverOBJ, PUNICODE_STRING pRegistryPath )
{

	NTSTATUS					ReturnStatus		= 0;
	int							i					= 0;

	ReturnStatus = STATUS_SUCCESS;

	for( i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++ )
	{
        pDriverOBJ->MajorFunction[i] = FilterDispatch;
    }

	pDriverOBJ->DriverExtension->AddDevice	= FilterAddRoutine;
	pDriverOBJ->DriverUnload				= FilterUnloadRoutine;

	pDriverOBJ->MajorFunction[ IRP_MJ_PNP ]				= FilterPnPRoutine;
	pDriverOBJ->MajorFunction[ IRP_MJ_INTERNAL_DEVICE_CONTROL ]	= FilterInternalDeviceIoControlRoutine;
	pDriverOBJ->MajorFunction[ IRP_MJ_DEVICE_CONTROL ]	= FilterDeviceIoControlRoutine;
	pDriverOBJ->MajorFunction[ IRP_MJ_POWER ]			= FilterPowerRoutine;
	
	return ReturnStatus;
}

NTSTATUS FilterAddRoutine( PDRIVER_OBJECT pDriverOBJ, PDEVICE_OBJECT pPhysicalOBJ )
{
	NTSTATUS				ReturnStatus		= 0;
	ULONG					ExtensionSize		= 0;
	PDEVICE_OBJECT			pFilterDeviceOBJ	= NULL;
	PDEVICE_OBJECT			pCustomDeviceOBJ	= NULL;
	PMOUSEFILTEREXTENSION	pDeviceExtension	= NULL;
	UNICODE_STRING				DeviceName;
	UNICODE_STRING				SymbolName;
	
	ExtensionSize = sizeof( MOUSEFILTEREXTENSION );

	RtlInitUnicodeString( &DeviceName, UNICODE_DEVICE_NAME );
	
	// 켱 α׷ ȭ ϰԵ DeviceObject غϴ
	ReturnStatus = IoCreateDevice( pDriverOBJ,
								   ExtensionSize,
								   &DeviceName,
								   FILE_DEVICE_MOUSE,
								   0,
								   FALSE,
								   &pCustomDeviceOBJ );
	if ( !NT_SUCCESS( ReturnStatus ) )
	{
		//  ̸  DeviceObject Ѵٸ, ̻ ͸  ʴ´
		return ReturnStatus;
	}

	RtlInitUnicodeString( &SymbolName, UNICODE_SYMBOL_NAME );
	ReturnStatus = IoCreateSymbolicLink( &SymbolName, &DeviceName );
	// α׷   ֵ ɺũ մϴ

	// Filter  DeviceObject غϴ
	ReturnStatus = IoCreateDevice( pDriverOBJ,
								   ExtensionSize,
								   NULL,
								   FILE_DEVICE_MOUSE,
								   0,
								   FALSE,
								   &pFilterDeviceOBJ );

	if ( !NT_SUCCESS( ReturnStatus ) )
	{
		return ReturnStatus;
	}

	pDeviceExtension = ( PMOUSEFILTEREXTENSION ) pFilterDeviceOBJ->DeviceExtension;
	RtlZeroMemory( pDeviceExtension, ExtensionSize );

	pDeviceExtension->bIsFilterDeviceObject = TRUE;
	pDeviceExtension->u.pCustomDeviceObject = pCustomDeviceOBJ;

	pDeviceExtension->bMouseBlock = FALSE;
	pDeviceExtension->pUserEvent = NULL;
	pDeviceExtension->Self = pFilterDeviceOBJ;
	pDeviceExtension->pTopLevelOBJ = IoAttachDeviceToDeviceStack( pFilterDeviceOBJ, pPhysicalOBJ );

	if ( pDeviceExtension->pTopLevelOBJ != NULL )
	{
		pFilterDeviceOBJ->Flags |= (pDeviceExtension->pTopLevelOBJ->Flags & ( DO_BUFFERED_IO | DO_DIRECT_IO | DO_POWER_PAGABLE ) );
		pFilterDeviceOBJ->Flags &= ~DO_DEVICE_INITIALIZING;

		pFilterDeviceOBJ->Characteristics	= pDeviceExtension->pTopLevelOBJ->Characteristics;
		pFilterDeviceOBJ->DeviceType		= pDeviceExtension->pTopLevelOBJ->DeviceType;

		// pCustomDeviceObject  ʱȭ
		
		pCustomDeviceOBJ->Flags &= ~DO_DEVICE_INITIALIZING;
		pDeviceExtension = (PMOUSEFILTEREXTENSION)pCustomDeviceOBJ->DeviceExtension;
		pDeviceExtension->bIsFilterDeviceObject = FALSE;
		pDeviceExtension->u.pFilterDeviceObject = pFilterDeviceOBJ;

	} else
	{
		ReturnStatus = STATUS_UNSUCCESSFUL;
	}

	return ReturnStatus;
}

NTSTATUS FilterUnloadRoutine( PDRIVER_OBJECT pDriverOBJ )
{
	DbgPrint( "MouseFilter : FilterUnloadRoutine\n" );
	return STATUS_SUCCESS;
}

NTSTATUS FilterDispatch( PDEVICE_OBJECT pDeviceOBJ, PIRP pIrp )
{
	NTSTATUS				ReturnStatus		= 0;
	PMOUSEFILTEREXTENSION	pDeviceExtension	= NULL;
	PIO_STACK_LOCATION		pCurrentStack		= NULL;

	pDeviceExtension = ( PMOUSEFILTEREXTENSION ) pDeviceOBJ->DeviceExtension;
	pCurrentStack = IoGetCurrentIrpStackLocation( pIrp );

	if ( pDeviceExtension->bIsFilterDeviceObject == FALSE )
	{
		// α׷ ȭ ϴ CustomDeviceObject Ȯմϴ
		pDeviceExtension = (PMOUSEFILTEREXTENSION)((pDeviceExtension->u.pFilterDeviceObject)->DeviceExtension);

		switch( pCurrentStack->MajorFunction )
		{
			case IRP_MJ_CLEANUP:
			{
				if ( pDeviceExtension->pUserEvent != NULL )
				{
					ObDereferenceObject( pDeviceExtension->pUserEvent );
					pDeviceExtension->pUserEvent = NULL;
				}
			}
		}
		pIrp->IoStatus.Status = STATUS_SUCCESS;
		pIrp->IoStatus.Information = 0;
		IoCompleteRequest( pIrp, IO_NO_INCREMENT );
		return STATUS_SUCCESS;
	}

	IoSkipCurrentIrpStackLocation( pIrp );
	ReturnStatus = IoCallDriver( pDeviceExtension->pTopLevelOBJ, pIrp );

	return ReturnStatus;
}

NTSTATUS FilterPnPRoutine( PDEVICE_OBJECT pDeviceOBJ, PIRP pIrp )
{
	NTSTATUS				ReturnStatus		= 0;
	PIO_STACK_LOCATION		pCurrentStack		= NULL;
	PMOUSEFILTEREXTENSION	pDeviceExtension	= NULL;
	UNICODE_STRING				SymbolName;

	pCurrentStack		= IoGetCurrentIrpStackLocation( pIrp );
	pDeviceExtension	= ( PMOUSEFILTEREXTENSION ) pDeviceOBJ->DeviceExtension;

	switch( pCurrentStack->MinorFunction )
	{
		case IRP_MN_REMOVE_DEVICE:
		{
			RtlInitUnicodeString( &SymbolName, UNICODE_SYMBOL_NAME );
			IoDeleteSymbolicLink( &SymbolName );

			IoDeleteDevice( pDeviceExtension->u.pCustomDeviceObject );

			IoDetachDevice( pDeviceExtension->pTopLevelOBJ );
			IoDeleteDevice( pDeviceOBJ );			
			break;
		}
	}

	IoSkipCurrentIrpStackLocation( pIrp );
	ReturnStatus = IoCallDriver( pDeviceExtension->pTopLevelOBJ, pIrp );

	return ReturnStatus;
}

VOID
FilterServiceCallback(
    PDEVICE_OBJECT DeviceObject,
    PMOUSE_INPUT_DATA InputDataStart,
    PMOUSE_INPUT_DATA InputDataEnd,
    OUT PULONG InputDataConsumed
    )
{
	PMOUSEFILTEREXTENSION		pDeviceExtension	= NULL;
	PMOUSE_INPUT_DATA pInputData;
	ULONG Button; 

    pDeviceExtension = (PMOUSEFILTEREXTENSION) DeviceObject->DeviceExtension;

	if( pDeviceExtension->bMouseBlock == TRUE )
	{
		for( pInputData = (PMOUSE_INPUT_DATA)InputDataStart; pInputData <= (PMOUSE_INPUT_DATA)InputDataEnd; pInputData++ )
		{
			Button = pInputData->Buttons;
			if( pInputData->Buttons & MOUSE_LEFT_BUTTON_DOWN )
			{
				Button &= ~MOUSE_LEFT_BUTTON_DOWN;
				Button |= MOUSE_RIGHT_BUTTON_DOWN;
			}
			if( pInputData->Buttons & MOUSE_LEFT_BUTTON_UP )
			{
				Button &= ~MOUSE_LEFT_BUTTON_UP;
				Button |= MOUSE_RIGHT_BUTTON_UP;
			}
			if( pInputData->Buttons & MOUSE_RIGHT_BUTTON_DOWN )
			{
				Button &= ~MOUSE_RIGHT_BUTTON_DOWN;
				Button |= MOUSE_LEFT_BUTTON_DOWN;
			}
			if( pInputData->Buttons & MOUSE_RIGHT_BUTTON_UP )
			{
				Button &= ~MOUSE_RIGHT_BUTTON_UP;
				Button |= MOUSE_LEFT_BUTTON_UP;
			}
			pInputData->Buttons = Button;

			if( pInputData->Buttons & MOUSE_MIDDLE_BUTTON_UP )
			{
				if( pDeviceExtension->pUserEvent )
					KeSetEvent( pDeviceExtension->pUserEvent, 0, FALSE );
			}

		}
	}

    (*(PSERVICE_CALLBACK_ROUTINE) pDeviceExtension->UpperConnectData.ClassService)(
        pDeviceExtension->UpperConnectData.ClassDeviceObject,
        InputDataStart,
        InputDataEnd,
        InputDataConsumed
        );
}

NTSTATUS FilterInternalDeviceIoControlRoutine( PDEVICE_OBJECT pDeviceOBJ, PIRP pIrp )
{
	NTSTATUS					ReturnStatus		= 0;
	PMOUSEFILTEREXTENSION		pDeviceExtension	= NULL;
	PIO_STACK_LOCATION			pCurrentStack		= NULL;
    PCONNECT_DATA               connectData;

	pDeviceExtension	= ( PMOUSEFILTEREXTENSION ) pDeviceOBJ->DeviceExtension;
	pCurrentStack		= IoGetCurrentIrpStackLocation( pIrp );

    switch (pCurrentStack->Parameters.DeviceIoControl.IoControlCode) {
	    case IOCTL_INTERNAL_MOUSE_CONNECT:
			if (pDeviceExtension->UpperConnectData.ClassService != NULL) {
				ReturnStatus = STATUS_SHARING_VIOLATION;
				break;
			}
			else if (pCurrentStack->Parameters.DeviceIoControl.InputBufferLength <
					sizeof(CONNECT_DATA)) {
				ReturnStatus = STATUS_INVALID_PARAMETER;
				break;
			}
			connectData = ((PCONNECT_DATA)
				(pCurrentStack->Parameters.DeviceIoControl.Type3InputBuffer));
			pDeviceExtension->UpperConnectData = *connectData;
			connectData->ClassDeviceObject = pDeviceExtension->Self;
			connectData->ClassService = FilterServiceCallback;
			break;
	}

    if (!NT_SUCCESS(ReturnStatus)) {
        pIrp->IoStatus.Status = ReturnStatus;
        pIrp->IoStatus.Information = 0;
        IoCompleteRequest(pIrp, IO_NO_INCREMENT);

        return ReturnStatus;
    }

	IoSkipCurrentIrpStackLocation( pIrp );
	ReturnStatus = IoCallDriver( pDeviceExtension->pTopLevelOBJ, pIrp );

	return ReturnStatus;
}

NTSTATUS FilterDeviceIoControlRoutine( PDEVICE_OBJECT pDeviceOBJ, PIRP pIrp )
{
	NTSTATUS					ReturnStatus		= 0;
	PMOUSEFILTEREXTENSION		pDeviceExtension	= NULL;
	PIO_STACK_LOCATION			pCurrentStack		= NULL;

	pDeviceExtension	= ( PMOUSEFILTEREXTENSION ) pDeviceOBJ->DeviceExtension;
	pCurrentStack		= IoGetCurrentIrpStackLocation( pIrp );

	if ( pDeviceExtension->bIsFilterDeviceObject == FALSE )
	{ 
		// α׷ ȭ ϴ CustomDeviceObject Ȯմϴ
		pDeviceExtension = (PMOUSEFILTEREXTENSION)((pDeviceExtension->u.pFilterDeviceObject)->DeviceExtension);
		switch( pCurrentStack->Parameters.DeviceIoControl.IoControlCode )
		{
			case IOCTL_FILTER_START :
			{
				pDeviceExtension->bMouseBlock = TRUE;
				break;
			}

			case IOCTL_FILTER_STOP :
			{
				pDeviceExtension->bMouseBlock = FALSE;
				break;
			}

			case IOCTL_EVENT_ADDR :
			{
				HANDLE						UserEventHandle		= NULL;

				UserEventHandle = *( HANDLE* )pIrp->AssociatedIrp.SystemBuffer;

				if( pDeviceExtension->pUserEvent )
					ObDereferenceObject( pDeviceExtension->pUserEvent );

				pDeviceExtension->pUserEvent = NULL;

				ReturnStatus = ObReferenceObjectByHandle( UserEventHandle,
														  EVENT_ALL_ACCESS,
														  *ExEventObjectType,
														  UserMode,
														  &( pDeviceExtension->pUserEvent ),
														  NULL );

				if ( !NT_SUCCESS( ReturnStatus ) )
				{
					pIrp->IoStatus.Information	= 0;
					pIrp->IoStatus.Status		= ReturnStatus;
					IoCompleteRequest( pIrp, IO_NO_INCREMENT );

					return ReturnStatus;
				}
				
				pIrp->IoStatus.Information	= pCurrentStack->Parameters.DeviceIoControl.InputBufferLength;
				pIrp->IoStatus.Status		= ReturnStatus;
				IoCompleteRequest( pIrp, IO_NO_INCREMENT );

				return ReturnStatus;
			}

			default:
			{
				break;
			}
		}

		pIrp->IoStatus.Status		= STATUS_SUCCESS;
		pIrp->IoStatus.Information	= 0;
		IoCompleteRequest( pIrp, IO_NO_INCREMENT );

		return STATUS_SUCCESS;
	}
	
	// Filter Ǵ DeviceObject Skipå ϴ
	IoSkipCurrentIrpStackLocation( pIrp );
	ReturnStatus = IoCallDriver( pDeviceExtension->pTopLevelOBJ, pIrp );

	return ReturnStatus;
}


NTSTATUS FilterPowerRoutine( PDEVICE_OBJECT pDeviceOBJ, PIRP pIrp )
{
	NTSTATUS				ReturnStatus		= 0;
	PMOUSEFILTEREXTENSION	pDeviceExtension	= NULL;

	pDeviceExtension	= ( PMOUSEFILTEREXTENSION ) pDeviceOBJ->DeviceExtension;

	PoStartNextPowerIrp( pIrp );
	IoSkipCurrentIrpStackLocation( pIrp );
	ReturnStatus = PoCallDriver( pDeviceExtension->pTopLevelOBJ, pIrp );

	return ReturnStatus;
}

